module jp.ku6.bitmap.bmp;

public import jp.ku6.bitmap.core;

import std.stream;

/++
 + BMPアクセスクラス
 +/
class MicrosoftWindowsBitmapImageFile : Bitmap
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  filePath = ファイルパス
	 +/
	this(string filePath)
	{
		filePath_ = filePath;
	}

	@property
	{
		uint width()
		{
			return width_;
		}

		uint height()
		{
			return height_;
		}
	}

	bool open()
	{
		// ファイルをオープン
		fileStream_ = new BufferedFile(filePath_);

		// ヘッダ情報を読み出し
		ubyte[54] buffer;
		fileStream_.read(buffer);

		// ファイル先頭のBMを確認
		if (buffer[0] != 0x42 || buffer[1] != 0x4D)
			goto OPEN_FAILED;

		// BITMAPFILEHEADERサイズが54か確認
		if (buffer[10] != 0x36 || buffer[11] != 0x00 || buffer[12] != 0x00 || buffer[13] != 0x00)
			goto OPEN_FAILED;

		// BITMAPINFOHEADERイズが40か確認
		if (buffer[14] != 0x28 || buffer[15] != 0x00 || buffer[16] != 0x00 || buffer[17] != 0x00)
			goto OPEN_FAILED;

		// 幅を取得
		width_ = buffer[18] | (buffer[19] << 8) | (buffer[20] << 16) | (buffer[21] << 24);

		// 高さを取得
		height_ = buffer[22] | (buffer[23] << 8) | (buffer[24] << 16) | (buffer[25] << 24);

		// プレーン数を確認
		if (buffer[26] != 0x01 || buffer[27] != 0x00)
			goto OPEN_FAILED;

		// ビット深度を取得
		bitPerPixel_ = buffer[28];
		if (bitPerPixel_ != 24)
			goto OPEN_FAILED;

		return true;

// 失敗時にはクローズしてから終わる
OPEN_FAILED:

		fileStream_.close();

		return false;
	}

	void close()
	{
		fileStream_.close();
	}

	uint[] readAsTrueColor()
	{
		uint[] destination = new uint[width_ * height_];
		return readAsTrueColor(destination, width_, height_);
	}

	uint[] readAsTrueColor(uint[] destination, uint destinationWidth, uint destinationHeight)
	{
		// 展開先バッファに展開できるサイズがあるか確認
		if (destination is null || destination.length == 0 || destinationWidth == 0 || destinationHeight == 0 || destination.length < destinationWidth * destinationHeight || destinationWidth < width_ || destinationHeight < height_)
			return null;

		// ファイルから変換元データ等をロード
		uint pitch = ((width_ * bitPerPixel_ / 8) + 3) & ~3;

		// データ読み込み領域を確保
		ubyte[] source = new ubyte[pitch * height_];

		// ヘッダの直後から読み込みを開始
		fileStream_.seek(54, SeekPos.Set);

		// ピクセルデータを読み込み
		fileStream_.read(source);

		// 上下反転しつつTrueColorにして転送
		for (uint y; y < height_; ++y)
		{
			uint* writeCursor = &destination[destinationWidth * y];
			ubyte* readCursor = &source[pitch * (height_ - y - 1)];
			for (uint x; x < width_; ++x)
				writeCursor[x] = (readCursor[x * 3] << 16) | (readCursor[x * 3 + 1] << 8) | readCursor[x * 3 + 2] | 0xff000000;
		}

		return destination;
	}

private:

	// ファイルパス
	string filePath_;

	// ファイルストリーム
	BufferedFile fileStream_;

	// 1ピクセルあたりのビット数
	ubyte bitPerPixel_;

	// 画像の幅
	uint width_;

	// 画像の高さ
	uint height_;
}
